import dataclasses as dc
import operator
import typing as ta

from omlish import marshal as msh


T = ta.TypeVar('T')


##


@dc.dataclass(frozen=True)
class Namespace:
    key: int
    case: str  # 'first-letter' | 'case-sensitive'
    text: str


#


@dc.dataclass(frozen=True)
class Namespaces:
    namespaces: ta.Sequence[Namespace] | None = None


#


@dc.dataclass(frozen=True)
class SiteInfo:
    sitename: str
    dbname: str
    base: str
    generator: str
    case: str  # 'first-letter' | 'case-sensitive'
    namespaces: Namespaces


#


@dc.dataclass(frozen=True)
class Redirect:
    title: str


#


@dc.dataclass(frozen=True)
@msh.update_fields_metadata([
    'username',
    'id',
    'ip',
    'deleted',
], omit_if=operator.not_)
class Contributor:
    username: str | None = None
    id: int | None = None
    ip: str | None = None
    deleted: bool | None = None


#


@dc.dataclass(frozen=True)
class RevisionText:
    bytes: int
    sha1: str
    text: str


#


@dc.dataclass(frozen=True)
@msh.update_fields_metadata([
    'parentid',
    'timestamp',
    'contributors',
    'minor',
    'comment',
    'origin',
    'model',
    'format',
    'text',
    'sha1',
], omit_if=operator.not_)
class Revision:
    id: int
    parentid: int | None = None
    timestamp: str | None = None  # ISO8601
    contributors: ta.Sequence[Contributor] | None = None
    minor: bool = False
    comment: str | None = None
    origin: int | None = None
    model: str | None = None
    format: str | None = None
    text: RevisionText | None = None
    sha1: str | None = None


#


@dc.dataclass(frozen=True)
@msh.update_fields_metadata([
    'timestamp',
    'contributor',
    'comment',
    'filename',
    'src',
    'size',
], omit_if=operator.not_)
class Upload:
    timestamp: str | None = None
    contributor: str | None = None
    comment: str | None = None
    filename: str | None = None
    src: str | None = None
    size: str | None = None


#


@dc.dataclass(frozen=True)
@msh.update_fields_metadata([
    'redirect',
    'restrictions',
    'revisions',
    'uploads',
], omit_if=operator.not_)
class Page:
    title: str
    ns: int
    id: int
    redirect: Redirect | None = None
    restrictions: str | None = None
    revisions: ta.Sequence[Revision] | None = None
    uploads: ta.Sequence[Upload] | None = None


#


@dc.dataclass(frozen=True)
class MediaWiki:
    siteinfo: SiteInfo | None = None
    pages: ta.Sequence[Page] | None = None



import typing as ta

from omlish import check
from omlish import lang

from ..chat.messages import AiMessage
from ..chat.messages import Message
from ..chat.messages import SystemMessage
from ..chat.messages import UserMessage
from ..chat.models import AiChoice
from ..chat.models import ChatModel
from ..chat.models import ChatRequest
from ..chat.models import ChatResponse


if ta.TYPE_CHECKING:
    import mlx.nn
    import mlx_lm.utils
else:
    mlx = lang.proxy_import('mlx', extras=['nn'])
    mlx_lm = lang.proxy_import('mlx_lm', extras=['utils'])


class MlxlmChatModel(ChatModel):
    DEFAULT_MODEL = (
        # 'mlx-community/DeepSeek-Coder-V2-Lite-Instruct-8bit'
        'mlx-community/Llama-3.3-70B-Instruct-4bit'
        # 'mlx-community/Llama-3.3-70B-Instruct-6bit'
        # 'mlx-community/Llama-3.3-70B-Instruct-8bit'
        # 'mlx-community/Mixtral-8x7B-Instruct-v0.1'
        # 'mlx-community/QwQ-32B-Preview-8bit'
        # 'mlx-community/QwQ-32B-Preview-bf16'
        # 'mlx-community/Qwen2.5-0.5B-4bit'
        # 'mlx-community/Qwen2.5-32B-Instruct-8bit'
        # 'mlx-community/Qwen2.5-Coder-32B-Instruct-8bit'
        # 'mlx-community/mamba-2.8b-hf-f16'
    )

    def __init__(
            self,
            model: str = DEFAULT_MODEL,
    ) -> None:
        super().__init__()
        self._model = model

    ROLES_MAP: ta.ClassVar[ta.Mapping[type[Message], str]] = {
        SystemMessage: 'system',
        UserMessage: 'user',
        AiMessage: 'assistant',
    }

    def _get_msg_content(self, m: Message) -> str | None:
        if isinstance(m, (SystemMessage, AiMessage)):
            return m.s

        elif isinstance(m, UserMessage):
            return check.isinstance(m.c, str)

        else:
            raise TypeError(m)

    class _LoadedModel(ta.NamedTuple):
        model: 'mlx.nn.Module'
        tokenizer: 'mlx_lm.utils.TokenizerWrapper'

    @lang.cached_function(transient=True)
    def _load_model(self) -> _LoadedModel:
        model, tokenizer = mlx_lm.load(self._model)
        return self._LoadedModel(model, tokenizer)

    def invoke(self, request: ChatRequest) -> ChatResponse:
        model, tokenizer = self._load_model()

        if not (
                hasattr(tokenizer, 'apply_chat_template') and
                tokenizer.chat_template is not None
        ):
            raise RuntimeError(tokenizer)

        prompt = tokenizer.apply_chat_template(
            [
                dict(
                    role=self.ROLES_MAP[type(m)],
                    content=self._get_msg_content(m),
                )
                for m in request.v
            ],
            tokenize=False,
            add_generation_prompt=True,
        )

        response = mlx_lm.generate(
            model,
            tokenizer,
            prompt=prompt,
            # verbose=True,
        )

        return ChatResponse(v=[
            AiChoice(AiMessage(response))  # noqa
        ])



from omlish import dataclasses as dc
from omlish import lang

from ..json import JsonSchema
from ..options import UniqueOption
from .options import ChatOption


##


class ResponseFormat(ChatOption, UniqueOption, lang.Abstract, lang.Sealed):
    pass


@dc.dataclass(frozen=True)
class TextResponseFormat(ResponseFormat, lang.Final):
    pass


TEXT_RESPONSE_FORMAT = TextResponseFormat()


@dc.dataclass(frozen=True)
class JsonResponseFormat(ResponseFormat, lang.Final):
    schema: JsonSchema | None = None


JSON_RESPONSE_FORMAT = JsonResponseFormat()



import typing as ta

from omlish import cached
from omlish import collections as col
from omlish import dataclasses as dc
from omlish import lang

from .dtypes import Dtype


##


@dc.dataclass(frozen=True)
class FieldRef(lang.Final):
    n: str = dc.xfield(check_type=True)

    def __str__(self) -> str:
        return f':{self.n}'


def f(o: str | FieldRef) -> FieldRef:
    if isinstance(o, FieldRef):
        return o
    return FieldRef(o)


def f_(o: str | FieldRef) -> str:
    if isinstance(o, str):
        return o
    elif isinstance(o, FieldRef):
        return o.n
    else:
        raise TypeError(o)


##


@dc.dataclass(frozen=True)
class DocField(lang.Final):
    name: str
    dtype: Dtype


@dc.dataclass(frozen=True)
class DocSchema(lang.Final):
    fields: ta.Sequence[DocField] = dc.xfield(validate=lambda v: all(isinstance(e, DocField) for e in v))

    @cached.property
    @dc.init
    def fields_by_name(self) -> ta.Mapping[str, DocField]:
        return col.make_map_by(lambda f: f.name, self.fields, strict=True)  # noqa

    def __getitem__(self, key: str | FieldRef) -> DocField:
        return self.fields_by_name[f_(key)]


@dc.dataclass(frozen=True)
class Doc(lang.Final):
    values: ta.Mapping[str, ta.Any]

    def __getitem__(self, key: str | FieldRef) -> DocField:
        return self.values[f_(key)]



"""
TODO:
 - debug launch
"""
import dataclasses as dc
import logging
import os.path

from omdev.home.paths import get_home_paths
from omlish import cached
from omlish import check
from omlish.daemons import spawning
from omlish.daemons.daemon import Daemon
from omlish.daemons.services import Service
from omlish.daemons.services import ServiceDaemon
from omlish.daemons.services import ServiceTarget
from omlish.logs import all as logs

from .server import McServer


log = logging.getLogger(__name__)


##


class McServerService(Service['McServerService.Config']):
    @dc.dataclass(frozen=True)
    class Config(Service.Config):
        server: McServer.Config = McServer.Config()

    def __init__(self, config: Config = Config()) -> None:
        super().__init__(config)

        self._server = McServer(config.server)

    @property
    def server(self) -> McServer:
        return self._server

    def _run(self) -> None:
        self._server.run()


##


def _multiprocessing_entrypoint(args: spawning.MultiprocessingSpawning.EntrypointArgs) -> None:
    svc = check.isinstance(check.isinstance(args.spawn.target, ServiceTarget).svc, McServerService)  # noqa

    if args.start_method == spawning.MultiprocessingSpawning.StartMethod.SPAWN:
        # TODO: RotatingFileHandler('server.log', maxBytes=1_000_000, backupCount=3)
        log_file = os.path.join(get_home_paths().log_dir, 'minichain', 'server.log')
        os.makedirs(os.path.dirname(log_file), exist_ok=True)

        logs.configure_standard_logging(
            'DEBUG',
            handler_factory=lambda: logging.FileHandler(log_file),
        )

    log.debug('%r', args)

    args.spawn.fn()


@cached.function(lock=True)
def mc_server_service_daemon() -> ServiceDaemon[McServerService, McServerService.Config]:
    pid_file = os.path.join(get_home_paths().run_dir, 'minichain', 'server.pid')

    return ServiceDaemon(
        McServerService.Config(
            McServer.Config(),
        ),

        Daemon.Config(
            # spawning=spawning.ThreadSpawning(),
            spawning=spawning.MultiprocessingSpawning(entrypoint=_multiprocessing_entrypoint),
            # spawning=spawning.ForkSpawning(),

            pid_file=pid_file,

            reparent_process=True,
        ),
    )



import os

from omlish import check
from omlish.formats import json
from omlish.http import all as http
from omlish.secrets.secrets import Secret

from ...vectors.embeddings import EmbeddingModel
from ...vectors.embeddings import EmbeddingRequest
from ...vectors.embeddings import EmbeddingResponse
from ...vectors.vectors import Vector


class OpenaiEmbeddingModel(EmbeddingModel):
    model = 'text-embedding-3-small'

    def __init__(
            self,
            *,
            api_key: Secret | str | None = None,
    ) -> None:
        super().__init__()
        self._api_key = Secret.of(api_key if api_key is not None else os.environ['OPENAI_API_KEY'])

    def invoke(self, request: EmbeddingRequest) -> EmbeddingResponse:
        raw_request = dict(
            model=self.model,
            input=check.isinstance(request.v, str),
        )

        raw_response = http.request(
            'https://api.openai.com/v1/embeddings',
            headers={
                http.consts.HEADER_CONTENT_TYPE: http.consts.CONTENT_TYPE_JSON,
                http.consts.HEADER_AUTH: http.consts.format_bearer_auth_header(check.not_none(self._api_key).reveal()),
            },
            data=json.dumps(raw_request).encode('utf-8'),
        )

        response = json.loads(check.not_none(raw_response.data).decode('utf-8'))

        return EmbeddingResponse(v=Vector(response['data'][0]['embedding']))



import typing as ta

from omlish import lang

from ..content.images import Image
from ..vectors.embeddings import EmbeddingModel
from ..vectors.embeddings import EmbeddingRequest
from ..vectors.embeddings import EmbeddingResponse
from ..vectors.vectors import Vector


if ta.TYPE_CHECKING:
    import sentence_transformers as st
else:
    st = lang.proxy_import('sentence_transformers')


class SentencetransformersEmbeddingModel(EmbeddingModel):
    model = 'clip-ViT-B-32'

    def invoke(self, request: EmbeddingRequest) -> EmbeddingResponse:
        mdl = st.SentenceTransformer(self.model)

        obj: ta.Any
        v = request.v
        if isinstance(v, str):
            obj = v
        elif isinstance(v, Image):
            obj = v.i
        else:
            raise TypeError(v)

        response = mdl.encode(obj)

        return EmbeddingResponse(v=Vector(response.tolist()))



import typing as ta

from omlish import lang


if ta.TYPE_CHECKING:
    import torch.backends
    import torch.nn
else:
    torch = lang.proxy_import('torch', extras=['backends', 'nn'])


def get_best_device() -> str | None:
    if torch.cuda.is_available():
        return 'cuda'
    elif torch.backends.mps.is_available():
        return 'mps'
    else:
        return None


CanMoveToDevice: ta.TypeAlias = ta.Union['torch.Tensor', 'torch.nn.Module']
CanMoveToDeviceT = ta.TypeVar('CanMoveToDeviceT', bound=CanMoveToDevice)


def to(o: CanMoveToDeviceT, device: str | None) -> CanMoveToDeviceT:
    if isinstance(o, torch.Tensor):
        if device is not None:
            return o.to(device)  # type: ignore
        else:
            return o  # type: ignore

    elif isinstance(o, torch.nn.Module):
        if device is not None:
            o.to(device)
        return o

    else:
        raise TypeError(o)



"""
https://developers.google.com/custom-search/vo1/reference/rest/v1/cse/list?apix=true
https://developers.google.com/custom-search/v1/reference/rest/v1/Search
https://google.aip.dev/127
"""
import typing as ta
import urllib.parse

from omlish import check
from omlish import dataclasses as dc
from omlish import lang
from omlish import marshal as msh
from omlish.formats import json
from omlish.http import all as http

from ...search import SearchHit
from ...search import SearchHits
from ...search import SearchRequest
from ...search import SearchResponse
from ...search import SearchService


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL, unknown_field='x')
class CseSearchResult(lang.Final):
    kind: str | None = None

    title: str | None = None
    html_title: str | None = None

    link: str | None = None
    display_link: str | None = None

    snippet: str | None = None
    html_snippet: str | None = None

    cache_id: str | None = None

    formatted_url: str | None = None
    html_formatted_url: str | None = None

    mime: str | None = None
    file_format: str | None = None

    x: ta.Mapping[str, ta.Any] | None = dc.field(default=None, repr=False)


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL, unknown_field='x')
class CseSearchInfo(lang.Final):
    search_time: float | None = None
    total_results: int | None = None

    x: ta.Mapping[str, ta.Any] | None = dc.field(default=None, repr=False)


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL, unknown_field='x')
class CseSearchResponse(lang.Final):
    kind: str | None = None
    info: CseSearchInfo | None = dc.xfield(None) | msh.update_field_metadata(name='searchInformation')
    items: ta.Sequence[CseSearchResult] | None = None

    x: ta.Mapping[str, ta.Any] | None = dc.field(default=None, repr=False)


class CseSearchService(SearchService):
    def __init__(
            self,
            cse_id: str | None = None,
            cse_api_key: str | None = None,
    ) -> None:
        super().__init__()
        self._cse_id = cse_id
        self._cse_api_key = cse_api_key

    def invoke(
            self,
            request: SearchRequest,
    ) -> SearchResponse:
        qs = urllib.parse.urlencode(dict(
            key=check.non_empty_str(self._cse_api_key),
            cx=check.non_empty_str(self._cse_id),
            q=request.v,
        ))
        resp = http.request(
            f'https://www.googleapis.com/customsearch/v1?{qs}',
        )
        out = check.not_none(resp.data)

        dct = json.loads(out.decode('utf-8'))
        res = msh.unmarshal(dct, CseSearchResponse)
        return SearchResponse(v=SearchHits(
            l=[
                SearchHit(
                    title=i.title,
                    url=i.link,
                    snippets=[i.snippet],
                )
                for i in res.items or ()
            ],
            total_results=res.info.total_results if res.info is not None else None,
        ))



"""
TODO:
 - multi
"""
import time
import typing as ta


T = ta.TypeVar('T')
U = ta.TypeVar('U')


class ProgressReporter(ta.Generic[T]):
    def __init__(
            self,
            *,
            fn: ta.Callable[[], T] | None = None,
            start: T | None = None,

            fmt: str | None = None,
            suffix: str | None = None,
            total: T | None = None,
            report_interval: T | None = None,
    ) -> None:
        super().__init__()

        if fn is not None and start is None:
            start = fn()
        elif start is not None and fn is None:
            pass
        else:
            raise Exception('Must specify either fn or start')
        self._fn = fn

        self._fmt = fmt
        self._suffix = suffix
        self._total_v = total
        self._report_interval_v = report_interval

        self._cur_v = start
        self._start_v = self._cur_v
        self._start_t = time.time()
        self._reported_v: T | None = None
        self._reported_t: float | None = None

    def update(self, cur: T | None = None) -> T:
        if self._fn is not None:
            if cur is not None:
                raise Exception('Must not specify cur when fn given')
            cur = self._fn()
        else:  # noqa
            if cur is None:
                raise Exception('Must specify cur when fn not given')
        self._cur_v = cur
        return cur

    class Progress(ta.NamedTuple, ta.Generic[U]):
        cur_v: U
        report_base_v: U
        report_elapsed_v: U

        cur_t: float
        report_base_t: float
        report_elapsed_t: float

    def progress(self) -> Progress[T]:
        return ProgressReporter.Progress(
            cur_v=(cur_v := self._cur_v),
            report_base_v=(base_v := self._reported_v if self._reported_v is not None else self._start_v),
            report_elapsed_v=cur_v - base_v,  # type: ignore

            cur_t=(cur_t := time.time()),
            report_base_t=(base_t := self._reported_t if self._reported_t is not None else self._start_t),
            report_elapsed_t=cur_t - base_t,
        )

    @property
    def should_report(self) -> bool:
        if self._report_interval_v is None:
            return False
        return self.progress().report_elapsed_v >= self._report_interval_v  # type: ignore

    def report(self) -> list[str]:
        prg = self.progress()

        def fmt_v(v: T) -> str:
            if self._fmt is None:
                return str(v)
            return f'{{:{self._fmt}}}'.format(v)  # noqa

        ps = []
        sfx = ' ' + self._suffix if self._suffix is not None else ''
        if self._total_v is not None:
            ps.extend([
                f'{fmt_v(prg.cur_v)} / {fmt_v(self._total_v)}' + sfx,
                f'{prg.cur_v / self._total_v * 100.:.02f} %',  # type: ignore
            ])
        else:
            ps.append(f'{fmt_v(self._cur_v)}' + sfx)
        ps.append(
            f'{fmt_v(type(prg.cur_v)(prg.report_elapsed_v / prg.report_elapsed_t))} '  # type: ignore
            f'{self._suffix if self._suffix is not None else ""}/s',
        )

        self._reported_v = prg.cur_v
        self._reported_t = prg.cur_t

        return ps
